from .. import main


@main.group()
def admin():
    '''
    Provides the admin API access.
    '''


def _attach_command():
    from . import (  # noqa
        agents,
        domains,
        etcd,
        groups,
        images,
        keypairs,
        license,
        resources,
        resource_policies,
        scaling_groups,
        sessions,
        storage,
        users,
        vfolders,
    )


_attach_command()



import asyncio
import collections
from decimal import Decimal
import getpass
import itertools
import json
import re
import secrets
import string
import sys
import traceback
from typing import (
    Mapping,
    Optional,
    Sequence,
    Tuple,
)
import uuid

import aiohttp
import click
from humanize import naturalsize
import tabulate as tabulate_mod
from tabulate import tabulate

from . import main
from .admin.sessions import session as cli_admin_session
from ..config import local_cache_path
from ..compat import asyncio_run, current_loop
from ..exceptions import BackendError, BackendAPIError
from ..session import Session, AsyncSession, is_legacy_server
from ..types import undefined
from .pretty import (
    print_info, print_wait, print_done, print_error, print_fail, print_warn,
    format_info,
)

_rx_range_key = re.compile(r'^[a-zA-Z_][a-zA-Z0-9_]*$')

tabulate_mod.PRESERVE_WHITESPACE = True


def drange(start: Decimal, stop: Decimal, num: int):
    """
    A simplified version of numpy.linspace with default options
    """
    delta = stop - start
    step = delta / (num - 1)
    yield from (start + step * Decimal(tick) for tick in range(0, num))


class RangeExprOptionType(click.ParamType):
    """
    Accepts a range expression which generates a range of values for a variable.

    Linear space range: "linspace:1,2,10" (start, stop, num) as in numpy.linspace
    Pythonic range: "range:1,10,2" (start, stop[, step]) as in Python's range
    Case range: "case:a,b,c" (comma-separated strings)
    """

    name = 'Range Expression'

    def convert(self, arg, param, ctx):
        key, value = arg.split('=', maxsplit=1)
        assert _rx_range_key.match(key), 'The key must be a valid slug string.'
        try:
            if value.startswith('case:'):
                return key, value[5:].split(',')
            elif value.startswith('linspace:'):
                start, stop, num = value[9:].split(',')
                return key, tuple(drange(Decimal(start), Decimal(stop), int(num)))
            elif value.startswith('range:'):
                range_args = map(int, value[6:].split(','))
                return key, tuple(range(*range_args))
            else:
                self.fail('Unrecognized range expression type', param, ctx)
        except ValueError as e:
            self.fail(str(e), param, ctx)


range_expr = RangeExprOptionType()


async def exec_loop(stdout, stderr, compute_session, mode, code, *, opts=None,
                    vprint_done=print_done, is_multi=False):
    """
    Fully streamed asynchronous version of the execute loop.
    """
    async with compute_session.stream_execute(code, mode=mode, opts=opts) as stream:
        async for result in stream:
            if result.type == aiohttp.WSMsgType.TEXT:
                result = json.loads(result.data)
            else:
                # future extension
                continue
            for rec in result.get('console', []):
                if rec[0] == 'stdout':
                    print(rec[1], end='', file=stdout)
                elif rec[0] == 'stderr':
                    print(rec[1], end='', file=stderr)
                else:
                    print('----- output record (type: {0}) -----'.format(rec[0]),
                          file=stdout)
                    print(rec[1], file=stdout)
                    print('----- end of record -----', file=stdout)
            stdout.flush()
            files = result.get('files', [])
            if files:
                print('--- generated files ---', file=stdout)
                for item in files:
                    print('{0}: {1}'.format(item['name'], item['url']), file=stdout)
                print('--- end of generated files ---', file=stdout)
            if result['status'] == 'clean-finished':
                exitCode = result.get('exitCode')
                msg = 'Clean finished. (exit code = {0})'.format(exitCode)
                if is_multi:
                    print(msg, file=stderr)
                vprint_done(msg)
            elif result['status'] == 'build-finished':
                exitCode = result.get('exitCode')
                msg = 'Build finished. (exit code = {0})'.format(exitCode)
                if is_multi:
                    print(msg, file=stderr)
                vprint_done(msg)
            elif result['status'] == 'finished':
                exitCode = result.get('exitCode')
                msg = 'Execution finished. (exit code = {0})'.format(exitCode)
                if is_multi:
                    print(msg, file=stderr)
                vprint_done(msg)
                break
            elif result['status'] == 'waiting-input':
                if result['options'].get('is_password', False):
                    code = getpass.getpass()
                else:
                    code = input()
                await stream.send_str(code)
            elif result['status'] == 'continued':
                pass


def exec_loop_sync(stdout, stderr, compute_session, mode, code, *, opts=None,
                   vprint_done=print_done):
    """
    Old synchronous polling version of the execute loop.
    """
    opts = opts if opts else {}
    run_id = None  # use server-assigned run ID
    while True:
        result = compute_session.execute(run_id, code, mode=mode, opts=opts)
        run_id = result['runId']
        opts.clear()  # used only once
        for rec in result['console']:
            if rec[0] == 'stdout':
                print(rec[1], end='', file=stdout)
            elif rec[0] == 'stderr':
                print(rec[1], end='', file=stderr)
            else:
                print('----- output record (type: {0}) -----'.format(rec[0]),
                      file=stdout)
                print(rec[1], file=stdout)
                print('----- end of record -----', file=stdout)
        stdout.flush()
        files = result.get('files', [])
        if files:
            print('--- generated files ---', file=stdout)
            for item in files:
                print('{0}: {1}'.format(item['name'], item['url']), file=stdout)
            print('--- end of generated files ---', file=stdout)
        if result['status'] == 'clean-finished':
            exitCode = result.get('exitCode')
            vprint_done('Clean finished. (exit code = {0}'.format(exitCode),
                        file=stdout)
            mode = 'continue'
            code = ''
        elif result['status'] == 'build-finished':
            exitCode = result.get('exitCode')
            vprint_done('Build finished. (exit code = {0})'.format(exitCode),
                        file=stdout)
            mode = 'continue'
            code = ''
        elif result['status'] == 'finished':
            exitCode = result.get('exitCode')
            vprint_done('Execution finished. (exit code = {0})'.format(exitCode),
                        file=stdout)
            break
        elif result['status'] == 'waiting-input':
            mode = 'input'
            if result['options'].get('is_password', False):
                code = getpass.getpass()
            else:
                code = input()
        elif result['status'] == 'continued':
            mode = 'continue'
            code = ''


async def exec_terminal(compute_session, *,
                        vprint_wait=print_wait, vprint_done=print_done):
    # async with compute_session.stream_pty() as stream: ...
    raise NotImplementedError


def _noop(*args, **kwargs):
    pass


def _format_stats(stats):
    formatted = []
    version = stats.pop('version', 1)
    stats.pop('status')
    if version == 1:
        stats.pop('precpu_used', None)
        stats.pop('precpu_system_used', None)
        stats.pop('cpu_system_used', None)
        for key, val in stats.items():
            if key.endswith('_size') or key.endswith('_bytes'):
                val = naturalsize(val, binary=True)
            elif key == 'cpu_used':
                key += '_msec'
                val = '{0:,}'.format(int(val))
            else:
                val = '{0:,}'.format(int(val))
            formatted.append((key, val))
    elif version == 2:
        max_integer_len = 0
        max_fraction_len = 0
        for key, metric in stats.items():
            unit = metric['unit_hint']
            if unit == 'bytes':
                val = metric.get('stats.max', metric['current'])
                val = naturalsize(val, binary=True)
                val, unit = val.rsplit(' ', maxsplit=1)
                val = '{:,}'.format(Decimal(val))
            elif unit == 'msec':
                val = '{:,}'.format(Decimal(metric['current']))
                unit = 'msec'
            elif unit == 'percent':
                val = metric['pct']
                unit = '%'
            else:
                val = metric['current']
                unit = ''
            if val is None:
                continue
            ip, _, fp = val.partition('.')
            max_integer_len = max(len(ip), max_integer_len)
            max_fraction_len = max(len(fp), max_fraction_len)
            formatted.append([key, val, unit])
        fstr_int_only = '{0:>' + str(max_integer_len) + '}'
        fstr_float = '{0:>' + str(max_integer_len) + '}.{1:<' + str(max_fraction_len) + '}'
        for item in formatted:
            ip, _, fp = item[1].partition('.')
            if fp == '':
                item[1] = fstr_int_only.format(ip) + ' ' * (max_fraction_len + 1)
            else:
                item[1] = fstr_float.format(ip, fp)
    else:
        print_warn('Unsupported statistics result version. Upgrade your client.')
    return tabulate(formatted)


def _prepare_resource_arg(resources):
    if resources:
        resources = {k: v for k, v in map(lambda s: s.split('=', 1), resources)}
    else:
        resources = {}  # use the defaults configured in the server
    return resources


def _prepare_env_arg(env):
    if env is not None:
        envs = {k: v for k, v in map(lambda s: s.split('=', 1), env)}
    else:
        envs = {}
    return envs


def _prepare_mount_arg(
    mount_args: Optional[Sequence[str]]
) -> Tuple[Sequence[str], Mapping[str, str]]:
    """
    Parse the list of mount arguments into a list of
    vfolder name and in-container mount path pairs.
    """
    mounts = set()
    mount_map = {}
    if mount_args is not None:
        for value in mount_args:
            if '=' in value:
                sp = value.split('=', maxsplit=1)
            elif ':' in value:  # docker-like volume mount mapping
                sp = value.split(':', maxsplit=1)
            else:
                sp = [value]
            mounts.add(sp[0])
            if len(sp) == 2:
                mount_map[sp[0]] = sp[1]
    return list(mounts), mount_map


@main.command()
@click.argument('image', type=str)
@click.argument('files', nargs=-1, type=click.Path())
@click.option('-t', '--name', '--client-token', metavar='NAME',
              help='Specify a human-readable session name. '
                   'If not set, a random hex string is used.')
# job scheduling options
@click.option('--type', metavar='SESSTYPE',
              type=click.Choice(['batch', 'interactive']),
              default='interactive',
              help='Either batch or interactive')
@click.option('--starts-at', metavar='STARTS_AT', type=str, default=None,
              help='Let session to be started at a specific or relative time.')
@click.option('--enqueue-only', is_flag=True,
              help='Enqueue the session and return immediately without waiting for its startup.')
@click.option('--max-wait', metavar='SECONDS', type=int, default=0,
              help='The maximum duration to wait until the session starts.')
@click.option('--no-reuse', is_flag=True,
              help='Do not reuse existing sessions but return an error.')
# query-mode options
@click.option('-c', '--code', metavar='CODE',
              help='The code snippet as a single string')
@click.option('--terminal', is_flag=True,
              help='Connect to the terminal-type compute_session.')
# batch-mode options
@click.option('--clean', metavar='CMD',
              help='Custom shell command for cleaning up the base directory')
@click.option('--build', metavar='CMD',
              help='Custom shell command for building the given files')
@click.option('--exec', metavar='CMD',
              help='Custom shell command for executing the given files')
@click.option('--basedir', metavar='PATH', type=click.Path(), default=None,
              help='Base directory path of uploaded files. '
                   'All uploaded files must reside inside this directory.')
# execution environment
@click.option('-e', '--env', metavar='KEY=VAL', type=str, multiple=True,
              help='Environment variable (may appear multiple times)')
# extra options
@click.option('--bootstrap-script', metavar='PATH', type=click.File('r'), default=None,
              help='A user-defined script to execute on startup.')
@click.option('--rm', is_flag=True,
              help='Terminate the session immediately after running '
                   'the given code or files')
@click.option('-s', '--stats', is_flag=True,
              help='Show resource usage statistics after termination '
                   '(only works if "--rm" is given)')
@click.option('--tag', type=str, default=None,
              help='User-defined tag string to annotate sessions.')
@click.option('-q', '--quiet', is_flag=True,
              help='Hide execution details but show only the compute_session outputs.')
# experiment support
@click.option('--env-range', metavar='RANGE_EXPR', multiple=True,
              type=range_expr, help='Range expression for environment variable.')
@click.option('--build-range', metavar='RANGE_EXPR', multiple=True,
              type=range_expr, help='Range expression for execution arguments.')
@click.option('--exec-range', metavar='RANGE_EXPR', multiple=True, type=range_expr,
              help='Range expression for execution arguments.')
@click.option('--max-parallel', metavar='NUM', type=int, default=2,
              help='The maximum number of parallel sessions.')
# resource spec
@click.option('-v', '--volume', '-m', '--mount', 'mount',
              metavar='NAME[=PATH]', type=str, multiple=True,
              help='User-owned virtual folder names to mount. '
                   'If path is not provided, virtual folder will be mounted under /home/work. '
                   'All virtual folders can only be mounted under /home/work. ')
@click.option('--scaling-group', '--sgroup', type=str, default=None,
              help='The scaling group to execute session. If not specified, '
                   'all available scaling groups are included in the scheduling.')
@click.option('-r', '--resources', '--resource', metavar='KEY=VAL', type=str, multiple=True,
              help='Set computation resources '
                   '(e.g: -r cpu=2 -r mem=256 -r cuda.device=1)')
@click.option('--cluster-size', metavar='NUMBER', type=int, default=1,
              help='The size of cluster in number of containers.')
@click.option('--cluster-mode', metavar='MODE',
              type=click.Choice(['single-node', 'multi-node']), default='single-node',
              help='The mode of clustering.')
@click.option('--resource-opts', metavar='KEY=VAL', type=str, multiple=True,
              help='Resource options for creating compute session. '
                   '(e.g: shmem=64m)')
# resource grouping
@click.option('-d', '--domain', metavar='DOMAIN_NAME', default=None,
              help='Domain name where the session will be spawned. '
                   'If not specified, config\'s domain name will be used.')
@click.option('-g', '--group', metavar='GROUP_NAME', default=None,
              help='Group name where the session is spawned. '
                   'User should be a member of the group to execute the code.')
@click.option('--preopen',  default=None,
              help='Pre-open service ports')
def run(image, files, name,                                 # base args
        type, starts_at, enqueue_only, max_wait, no_reuse,  # job scheduling options
        code, terminal,                                     # query-mode options
        clean, build, exec, basedir,                        # batch-mode options
        env,                                                # execution environment
        bootstrap_script, rm, stats, tag, quiet,            # extra options
        env_range, build_range, exec_range, max_parallel,   # experiment support
        mount, scaling_group, resources,                    # resource spec
        cluster_size, cluster_mode,
        resource_opts,
        domain, group, preopen):                            # resource grouping
    """
    Run the given code snippet or files in a session.
    Depending on the session ID you give (default is random),
    it may reuse an existing session or create a new one.

    \b
    IMAGE: The name (and version/platform tags appended after a colon) of session
          runtime or programming language.')
    FILES: The code file(s). Can be added multiple times.
    """
    if quiet:
        vprint_info = vprint_wait = vprint_done = _noop
    else:
        vprint_info = print_info
        vprint_wait = print_wait
        vprint_done = print_done
    if files and code:
        print('You can run only either source files or command-line '
              'code snippet.', file=sys.stderr)
        sys.exit(1)
    if not files and not code:
        print('You should provide the command-line code snippet using '
              '"-c" option if run without files.', file=sys.stderr)
        sys.exit(1)

    envs = _prepare_env_arg(env)
    resources = _prepare_resource_arg(resources)
    resource_opts = _prepare_resource_arg(resource_opts)
    mount, mount_map = _prepare_mount_arg(mount)

    if env_range is None: env_range = []      # noqa
    if build_range is None: build_range = []  # noqa
    if exec_range is None: exec_range = []    # noqa

    env_ranges = {v: r for v, r in env_range}
    build_ranges = {v: r for v, r in build_range}
    exec_ranges = {v: r for v, r in exec_range}

    env_var_maps = [dict(zip(env_ranges.keys(), values))
                    for values in itertools.product(*env_ranges.values())]
    build_var_maps = [dict(zip(build_ranges.keys(), values))
                      for values in itertools.product(*build_ranges.values())]
    exec_var_maps = [dict(zip(exec_ranges.keys(), values))
                     for values in itertools.product(*exec_ranges.values())]
    case_set = collections.OrderedDict()
    vmaps_product = itertools.product(env_var_maps, build_var_maps, exec_var_maps)
    build_template = string.Template(build)
    exec_template = string.Template(exec)
    env_templates = {k: string.Template(v) for k, v in envs.items()}
    preopen_ports = [] if preopen is None else list(map(int, preopen.split(',')))
    for env_vmap, build_vmap, exec_vmap in vmaps_product:
        interpolated_envs = tuple((k, vt.substitute(env_vmap))
                                  for k, vt in env_templates.items())
        if build:
            interpolated_build = build_template.substitute(build_vmap)
        else:
            interpolated_build = '*'
        if exec:
            interpolated_exec = exec_template.substitute(exec_vmap)
        else:
            interpolated_exec = '*'
        case_set[(interpolated_envs, interpolated_build, interpolated_exec)] = 1

    is_multi = (len(case_set) > 1)
    if is_multi:
        if max_parallel <= 0:
            print('The number maximum parallel sessions must be '
                  'a positive integer.', file=sys.stderr)
            sys.exit(1)
        if terminal:
            print('You cannot run multiple cases with terminal.', file=sys.stderr)
            sys.exit(1)
        if not quiet:
            vprint_info('Running multiple sessions for the following combinations:')
            for case in case_set.keys():
                pretty_env = ' '.join('{}={}'.format(item[0], item[1])
                                      for item in case[0])
                print('env = {!r}, build = {!r}, exec = {!r}'
                      .format(pretty_env, case[1], case[2]))

    def _run_legacy(session, idx, name, envs,
                    clean_cmd, build_cmd, exec_cmd):
        try:
            compute_session = session.ComputeSession.get_or_create(
                image,
                name=name,
                type_=type,
                enqueue_only=enqueue_only,
                max_wait=max_wait,
                no_reuse=no_reuse,
                cluster_size=cluster_size,
                cluster_mode=cluster_mode,
                mounts=mount,
                mount_map=mount_map,
                envs=envs,
                resources=resources,
                domain_name=domain,
                group_name=group,
                scaling_group=scaling_group,
                tag=tag)
        except Exception as e:
            print_error(e)
            sys.exit(1)
        if compute_session.status == 'PENDING':
            print_info('Session ID {0} is enqueued for scheduling.'
                       .format(name))
            return
        elif compute_session.status == 'SCHEDULED':
            print_info('Session ID {0} is scheduled and about to be started.'
                       .format(name))
            return
        elif compute_session.status == 'RUNNING':
            if compute_session.created:
                vprint_done(
                    '[{0}] Session {1} is ready (domain={2}, group={3}).'
                    .format(idx, compute_session.name,
                            compute_session.domain, compute_session.group))
            else:
                vprint_done('[{0}] Reusing session {1}...'.format(idx, compute_session.name))
        elif compute_session.status == 'TERMINATED':
            print_warn('Session ID {0} is already terminated.\n'
                       'This may be an error in the compute_session image.'
                       .format(name))
            return
        elif compute_session.status == 'TIMEOUT':
            print_info('Session ID {0} is still on the job queue.'
                       .format(name))
            return
        elif compute_session.status in ('ERROR', 'CANCELLED'):
            print_fail('Session ID {0} has an error during scheduling/startup or cancelled.'
                       .format(name))
            return

        try:
            if files:
                vprint_wait('[{0}] Uploading source files...'.format(idx))
                ret = compute_session.upload(files, basedir=basedir,
                                    show_progress=True)
                if ret.status // 100 != 2:
                    print_fail('[{0}] Uploading source files failed!'.format(idx))
                    print('{0}: {1}\n{2}'.format(
                        ret.status, ret.reason, ret.text()))
                    return
                vprint_done('[{0}] Uploading done.'.format(idx))
                opts = {
                    'clean': clean_cmd,
                    'build': build_cmd,
                    'exec': exec_cmd,
                }
                if not terminal:
                    exec_loop_sync(sys.stdout, sys.stderr, compute_session, 'batch', '',
                                   opts=opts,
                                   vprint_done=vprint_done)
            if terminal:
                raise NotImplementedError('Terminal access is not supported in '
                                          'the legacy synchronous mode.')
            if code:
                exec_loop_sync(sys.stdout, sys.stderr, compute_session, 'query', code,
                               vprint_done=vprint_done)
            vprint_done('[{0}] Execution finished.'.format(idx))
        except Exception as e:
            print_error(e)
            sys.exit(1)
        finally:
            if rm:
                vprint_wait('[{0}] Cleaning up the session...'.format(idx))
                ret = compute_session.destroy()
                vprint_done('[{0}] Cleaned up the session.'.format(idx))
                if stats:
                    _stats = ret.get('stats', None) if ret else None
                    if _stats:
                        print('[{0}] Statistics:\n{1}'
                              .format(idx, _format_stats(_stats)))
                    else:
                        print('[{0}] Statistics is not available.'.format(idx))

    async def _run(session, idx, name, envs,
                   clean_cmd, build_cmd, exec_cmd,
                   is_multi=False):
        try:
            compute_session = await session.ComputeSession.get_or_create(
                image,
                name=name,
                type_=type,
                starts_at=starts_at,
                enqueue_only=enqueue_only,
                max_wait=max_wait,
                no_reuse=no_reuse,
                cluster_size=cluster_size,
                cluster_mode=cluster_mode,
                mounts=mount,
                mount_map=mount_map,
                envs=envs,
                resources=resources,
                resource_opts=resource_opts,
                domain_name=domain,
                group_name=group,
                scaling_group=scaling_group,
                bootstrap_script=bootstrap_script.read() if bootstrap_script is not None else None,
                tag=tag,
                preopen_ports=preopen_ports)
        except Exception as e:
            print_fail('[{0}] {1}'.format(idx, e))
            return
        if compute_session.status == 'PENDING':
            print_info('Session ID {0} is enqueued for scheduling.'
                       .format(name))
            return
        elif compute_session.status == 'SCHEDULED':
            print_info('Session ID {0} is scheduled and about to be started.'
                       .format(name))
            return
        elif compute_session.status == 'RUNNING':
            if compute_session.created:
                vprint_done(
                    '[{0}] Session {1} is ready (domain={2}, group={3}).'
                    .format(idx, compute_session.name,
                            compute_session.domain, compute_session.group))
            else:
                vprint_done('[{0}] Reusing session {1}...'.format(idx, compute_session.name))
        elif compute_session.status == 'TERMINATED':
            print_warn('Session ID {0} is already terminated.\n'
                       'This may be an error in the compute_session image.'
                       .format(name))
            return
        elif compute_session.status == 'TIMEOUT':
            print_info('Session ID {0} is still on the job queue.'
                       .format(name))
            return
        elif compute_session.status in ('ERROR', 'CANCELLED'):
            print_fail('Session ID {0} has an error during scheduling/startup or cancelled.'
                       .format(name))
            return

        if not is_multi:
            stdout = sys.stdout
            stderr = sys.stderr
        else:
            log_dir = local_cache_path / 'client-logs'
            log_dir.mkdir(parents=True, exist_ok=True)
            stdout = open(log_dir / '{0}.stdout.log'.format(name),
                          'w', encoding='utf-8')
            stderr = open(log_dir / '{0}.stderr.log'.format(name),
                          'w', encoding='utf-8')

        try:
            def indexed_vprint_done(msg):
                vprint_done('[{0}] '.format(idx) + msg)
            if files:
                if not is_multi:
                    vprint_wait('[{0}] Uploading source files...'.format(idx))
                ret = await compute_session.upload(files, basedir=basedir,
                                          show_progress=not is_multi)
                if ret.status // 100 != 2:
                    print_fail('[{0}] Uploading source files failed!'.format(idx))
                    print('{0}: {1}\n{2}'.format(
                        ret.status, ret.reason, ret.text()), file=stderr)
                    raise RuntimeError('Uploading source files has failed!')
                if not is_multi:
                    vprint_done('[{0}] Uploading done.'.format(idx))
                opts = {
                    'clean': clean_cmd,
                    'build': build_cmd,
                    'exec': exec_cmd,
                }
                if not terminal:
                    await exec_loop(stdout, stderr, compute_session, 'batch', '',
                                    opts=opts,
                                    vprint_done=indexed_vprint_done,
                                    is_multi=is_multi)
            if terminal:
                await exec_terminal(compute_session)
                return
            if code:
                await exec_loop(stdout, stderr, compute_session, 'query', code,
                                vprint_done=indexed_vprint_done,
                                is_multi=is_multi)
        except BackendError as e:
            print_fail('[{0}] {1}'.format(idx, e))
            raise RuntimeError(e)
        except Exception as e:
            print_fail('[{0}] Execution failed!'.format(idx))
            traceback.print_exc()
            raise RuntimeError(e)
        finally:
            try:
                if rm:
                    if not is_multi:
                        vprint_wait('[{0}] Cleaning up the session...'.format(idx))
                    ret = await compute_session.destroy()
                    vprint_done('[{0}] Cleaned up the session.'.format(idx))
                    if stats:
                        _stats = ret.get('stats', None) if ret else None
                        if _stats:
                            stats_str = _format_stats(_stats)
                            print(format_info('[{0}] Statistics:'.format(idx)) +
                                  '\n{0}'.format(stats_str))
                            if is_multi:
                                print('Statistics:\n{0}'.format(stats_str),
                                      file=stderr)
                        else:
                            print_warn('[{0}] Statistics: unavailable.'.format(idx))
                            if is_multi:
                                print('Statistics: unavailable.', file=stderr)
            except Exception as e:
                print_fail('[{0}] Error while printing stats'.format(idx))
                traceback.print_exc()
                raise RuntimeError(e)
            finally:
                if is_multi:
                    stdout.close()
                    stderr.close()

    def _run_cases_legacy():
        if name is None:
            name_prefix = f'pysdk-{secrets.token_hex(5)}'
        else:
            name_prefix = name
        vprint_info('Session token prefix: {0}'.format(name_prefix))
        vprint_info('In the legacy mode, all cases will run serially!')
        with Session() as session:
            for idx, case in enumerate(case_set.keys()):
                if is_multi:
                    _name = '{0}-{1}'.format(name_prefix, idx)
                else:
                    _name = name_prefix
                envs = dict(case[0])
                clean_cmd = clean if clean else '*'
                build_cmd = case[1]
                exec_cmd = case[2]
                _run_legacy(session, idx, _name, envs,
                            clean_cmd, build_cmd, exec_cmd)

    async def _run_cases():
        loop = current_loop()
        if name is None:
            name_prefix = f'pysdk-{secrets.token_hex(5)}'
        else:
            name_prefix = name
        vprint_info('Session name prefix: {0}'.format(name_prefix))
        if is_multi:
            print_info('Check out the stdout/stderr logs stored in '
                       '~/.cache/backend.ai/client-logs directory.')
        async with AsyncSession() as session:
            tasks = []
            # TODO: limit max-parallelism using aiojobs
            for idx, case in enumerate(case_set.keys()):
                if is_multi:
                    _name = '{0}-{1}'.format(name_prefix, idx)
                else:
                    _name = name_prefix
                envs = dict(case[0])
                clean_cmd = clean if clean else '*'
                build_cmd = case[1]
                exec_cmd = case[2]
                t = loop.create_task(
                    _run(session, idx, _name, envs,
                         clean_cmd, build_cmd, exec_cmd,
                         is_multi=is_multi))
                tasks.append(t)
            results = await asyncio.gather(*tasks, return_exceptions=True)
            if any(map(lambda r: isinstance(r, Exception), results)):
                if is_multi:
                    print_fail('There were failed cases!')
                sys.exit(1)

    try:
        if is_legacy_server():
            _run_cases_legacy()
        else:
            asyncio_run(_run_cases())
    except Exception as e:
        print_fail('{0}'.format(e))


@main.command()
@click.argument('image')
@click.option('-t', '--name', '--client-token', metavar='NAME',
              help='Specify a human-readable session name. '
                   'If not set, a random hex string is used.')
@click.option('-o', '--owner', '--owner-access-key', metavar='ACCESS_KEY',
              help='Set the owner of the target session explicitly.')
# job scheduling options
@click.option('--type', metavar='SESSTYPE',
              type=click.Choice(['batch', 'interactive']),
              default='interactive',
              help='Either batch or interactive')
@click.option('--starts-at', metavar='STARTS_AT', type=str, default=None,
              help='Let session to be started at a specific or relative time.')
@click.option('-c', '--startup-command', metavar='COMMAND',
              help='Set the command to execute for batch-type sessions.')
@click.option('--enqueue-only', is_flag=True,
              help='Enqueue the session and return immediately without waiting for its startup.')
@click.option('--max-wait', metavar='SECONDS', type=int, default=0,
              help='The maximum duration to wait until the session starts.')
@click.option('--no-reuse', is_flag=True,
              help='Do not reuse existing sessions but return an error.')
# execution environment
@click.option('-e', '--env', metavar='KEY=VAL', type=str, multiple=True,
              help='Environment variable (may appear multiple times)')
# extra options
@click.option('--bootstrap-script', metavar='PATH', type=click.File('r'), default=None,
              help='A user-defined script to execute on startup.')
@click.option('--tag', type=str, default=None,
              help='User-defined tag string to annotate sessions.')
# resource spec
@click.option('-v', '--volume', '-m', '--mount', 'mount',
              metavar='NAME[=PATH]', type=str, multiple=True,
              help='User-owned virtual folder names to mount. '
                   'If path is not provided, virtual folder will be mounted under /home/work. '
                   'All virtual folders can only be mounted under /home/work. ')
@click.option('--scaling-group', '--sgroup', type=str, default=None,
              help='The scaling group to execute session. If not specified, '
                   'all available scaling groups are included in the scheduling.')
@click.option('-r', '--resources', metavar='KEY=VAL', type=str, multiple=True,
              help='Set computation resources used by the session '
                   '(e.g: -r cpu=2 -r mem=256 -r gpu=1).'
                   '1 slot of cpu/gpu represents 1 core. '
                   'The unit of mem(ory) is MiB.')
@click.option('--cluster-size', metavar='NUMBER', type=int, default=1,
              help='The size of cluster in number of containers.')
@click.option('--cluster-mode', metavar='MODE',
              type=click.Choice(['single-node', 'multi-node']), default='single-node',
              help='The mode of clustering.')
@click.option('--resource-opts', metavar='KEY=VAL', type=str, multiple=True,
              help='Resource options for creating compute session '
                   '(e.g: shmem=64m)')
# resource grouping
@click.option('-d', '--domain', metavar='DOMAIN_NAME', default=None,
              help='Domain name where the session will be spawned. '
                   'If not specified, config\'s domain name will be used.')
@click.option('-g', '--group', metavar='GROUP_NAME', default=None,
              help='Group name where the session is spawned. '
                   'User should be a member of the group to execute the code.')
@click.option('--preopen',  default=None,
              help='Pre-open service ports')
def start(image, name, owner,                                 # base args
          type, starts_at, startup_command, enqueue_only, max_wait, no_reuse,  # job scheduling options
          env,                                            # execution environment
          bootstrap_script, tag,                          # extra options
          mount, scaling_group, resources,                # resource spec
          cluster_size, cluster_mode,
          resource_opts,
          domain, group, preopen):                        # resource grouping
    """
    Prepare and start a single compute session without executing codes.
    You may use the created session to execute codes using the "run" command
    or connect to an application service provided by the session using the "app"
    command.


    \b
    IMAGE: The name (and version/platform tags appended after a colon) of session
           runtime or programming language.
    """
    if name is None:
        name = f'pysdk-{secrets.token_hex(5)}'
    else:
        name = name

    ######
    envs = _prepare_env_arg(env)
    resources = _prepare_resource_arg(resources)
    resource_opts = _prepare_resource_arg(resource_opts)
    mount, mount_map = _prepare_mount_arg(mount)
    preopen_ports = [] if preopen is None else list(map(int, preopen.split(',')))
    with Session() as session:
        try:
            compute_session = session.ComputeSession.get_or_create(
                image,
                name=name,
                type_=type,
                starts_at=starts_at,
                enqueue_only=enqueue_only,
                max_wait=max_wait,
                no_reuse=no_reuse,
                cluster_size=cluster_size,
                cluster_mode=cluster_mode,
                mounts=mount,
                mount_map=mount_map,
                envs=envs,
                startup_command=startup_command,
                resources=resources,
                resource_opts=resource_opts,
                owner_access_key=owner,
                domain_name=domain,
                group_name=group,
                scaling_group=scaling_group,
                bootstrap_script=bootstrap_script.read() if bootstrap_script is not None else None,
                tag=tag,
                preopen_ports=preopen_ports)
        except Exception as e:
            print_error(e)
            sys.exit(1)
        else:
            if compute_session.status == 'PENDING':
                print_info('Session ID {0} is enqueued for scheduling.'
                           .format(compute_session.id))
            elif compute_session.status == 'SCHEDULED':
                print_info('Session ID {0} is scheduled and about to be started.'
                           .format(compute_session.id))
                return
            elif compute_session.status == 'RUNNING':
                if compute_session.created:
                    print_info('Session ID {0} is created and ready.'
                               .format(compute_session.id))
                else:
                    print_info('Session ID {0} is already running and ready.'
                               .format(compute_session.id))
                if compute_session.service_ports:
                    print_info('This session provides the following app services: ' +
                               ', '.join(sport['name']
                                         for sport in compute_session.service_ports))
            elif compute_session.status == 'TERMINATED':
                print_warn('Session ID {0} is already terminated.\n'
                           'This may be an error in the compute_session image.'
                           .format(compute_session.id))
            elif compute_session.status == 'TIMEOUT':
                print_info('Session ID {0} is still on the job queue.'
                           .format(compute_session.id))
            elif compute_session.status in ('ERROR', 'CANCELLED'):
                print_fail('Session ID {0} has an error during scheduling/startup or cancelled.'
                           .format(compute_session.id))


@main.command()
@click.argument('template_id')
@click.option('-t', '--name', '--client-token', metavar='NAME',
              default=undefined,
              help='Specify a human-readable session name. '
                   'If not set, a random hex string is used.')
@click.option('-o', '--owner', '--owner-access-key', metavar='ACCESS_KEY',
              default=undefined,
              help='Set the owner of the target session explicitly.')
# job scheduling options
@click.option('--type', 'type_', metavar='SESSTYPE',
              type=click.Choice(['batch', 'interactive', undefined]),  # type: ignore
              default=undefined,
              help='Either batch or interactive')
@click.option('--starts_at', metavar='STARTS_AT', type=str, default=None,
              help='Let session to be started at a specific or relative time.')
@click.option('-i', '--image', default=undefined,
              help='Set compute_session image to run.')
@click.option('-c', '--startup-command', metavar='COMMAND',
              default=undefined,
              help='Set the command to execute for batch-type sessions.')
@click.option('--enqueue-only', is_flag=True,
              help='Enqueue the session and return immediately without waiting for its startup.')
@click.option('--max-wait', metavar='SECONDS', type=int,
              default=-1,
              help='The maximum duration to wait until the session starts.')
@click.option('--no-reuse', is_flag=True,
              help='Do not reuse existing sessions but return an error.')
# execution environment
@click.option('-e', '--env', metavar='KEY=VAL', type=str, multiple=True,
              help='Environment variable (may appear multiple times)')
# extra options
@click.option('--tag', type=str,
              default='$NODEF$',
              help='User-defined tag string to annotate sessions.')
# resource spec
@click.option('-m', '--mount', metavar='NAME[=PATH]', type=str, multiple=True,
              help='User-owned virtual folder names to mount. '
                   'If path is not provided, virtual folder will be mounted under /home/work. '
                   'All virtual folders can only be mounted under /home/work. ')
@click.option('--scaling-group', '--sgroup', type=str,
              default='$NODEF$',
              help='The scaling group to execute session. If not specified, '
                   'all available scaling groups are included in the scheduling.')
@click.option('-r', '--resources', metavar='KEY=VAL', type=str, multiple=True,
              help='Set computation resources used by the session '
                   '(e.g: -r cpu=2 -r mem=256 -r gpu=1).'
                   '1 slot of cpu/gpu represents 1 core. '
                   'The unit of mem(ory) is MiB.')
@click.option('--cluster-size', metavar='NUMBER', type=int,
              default=-1,
              help='The size of cluster in number of containers.')
@click.option('--resource-opts', metavar='KEY=VAL', type=str, multiple=True,
              help='Resource options for creating compute session '
                   '(e.g: shmem=64m)')
# resource grouping
@click.option('-d', '--domain', metavar='DOMAIN_NAME', default=None,
              help='Domain name where the session will be spawned. '
                   'If not specified, config\'s domain name will be used.')
@click.option('-g', '--group', metavar='GROUP_NAME', default=None,
              help='Group name where the session is spawned. '
                   'User should be a member of the group to execute the code.')
@click.option('--no-mount', is_flag=True,
              help='If specified, client.py will tell server not to mount '
                   'any vFolders specified at template,')
@click.option('--no-env', is_flag=True,
              help='If specified, client.py will tell server not to add '
                   'any environs specified at template,')
@click.option('--no-resource', is_flag=True,
              help='If specified, client.py will tell server not to add '
                   'any resource specified at template,')
def start_template(
    template_id, name, owner,        # base args
    type_, starts_at, image, startup_command, enqueue_only, max_wait, no_reuse,  # job scheduling options
    env,                                            # execution environment
    tag,                                            # extra options
    mount, scaling_group, resources, cluster_size,  # resource spec
    resource_opts,
    domain, group,                                  # resource grouping
    no_mount, no_env, no_resource,
):
    """
    Prepare and start a single compute session without executing codes.
    You may use the created session to execute codes using the "run" command
    or connect to an application service provided by the session using the "app"
    command.


    \b
    IMAGE: The name (and version/platform tags appended after a colon) of session
           runtime or programming language.
    """
    if name is undefined:
        name = f'pysdk-{secrets.token_hex(5)}'
    else:
        name = name

    if max_wait == -1:
        max_wait = undefined
    if tag == '$NODEF$':
        tag = undefined
    if scaling_group == '$NODEF$':
        scaling_group = undefined
    if cluster_size == -1:
        cluster_size = undefined

    ######

    envs = _prepare_env_arg(env) if len(env) > 0 or no_env else undefined
    resources = _prepare_resource_arg(resources) if len(resources) > 0 or no_resource else undefined
    resource_opts = (_prepare_resource_arg(resource_opts)
                     if len(resource_opts) > 0 or no_resource else undefined)
    mount, mount_map = (_prepare_mount_arg(mount)
                        if len(mount) > 0 or no_mount else (undefined, undefined))
    with Session() as session:
        try:
            compute_session = session.ComputeSession.create_from_template(
                template_id,
                image=image,
                name=name,
                type_=type_,
                starts_at=starts_at,
                enqueue_only=enqueue_only,
                max_wait=max_wait,
                no_reuse=no_reuse,
                cluster_size=cluster_size,
                mounts=mount,
                mount_map=mount_map,
                envs=envs,
                startup_command=startup_command,
                resources=resources,
                resource_opts=resource_opts,
                owner_access_key=owner,
                domain_name=domain,
                group_name=group,
                scaling_group=scaling_group,
                tag=tag)
        except Exception as e:
            print_error(e)
            sys.exit(1)
        else:
            if compute_session.status == 'PENDING':
                print_info('Session ID {0} is enqueued for scheduling.'
                           .format(name))
            elif compute_session.status == 'SCHEDULED':
                print_info('Session ID {0} is scheduled and about to be started.'
                           .format(name))
                return
            elif compute_session.status == 'RUNNING':
                if compute_session.created:
                    print_info('Session ID {0} is created and ready.'
                               .format(name))
                else:
                    print_info('Session ID {0} is already running and ready.'
                               .format(name))
                if compute_session.service_ports:
                    print_info('This session provides the following app services: ' +
                               ', '.join(sport['name']
                                         for sport in compute_session.service_ports))
            elif compute_session.status == 'TERMINATED':
                print_warn('Session ID {0} is already terminated.\n'
                           'This may be an error in the compute_session image.'
                           .format(name))
            elif compute_session.status == 'TIMEOUT':
                print_info('Session ID {0} is still on the job queue.'
                           .format(name))
            elif compute_session.status in ('ERROR', 'CANCELLED'):
                print_fail('Session ID {0} has an error during scheduling/startup or cancelled.'
                           .format(name))


@main.command(aliases=['rm', 'kill'])
@click.argument('session_names', metavar='SESSID', nargs=-1)
@click.option('-f', '--forced', is_flag=True,
              help='Force-terminate the errored sessions (only allowed for admins)')
@click.option('-o', '--owner', '--owner-access-key', metavar='ACCESS_KEY',
              help='Specify the owner of the target session explicitly.')
@click.option('-s', '--stats', is_flag=True,
              help='Show resource usage statistics after termination')
def terminate(session_names, forced, owner, stats):
    """
    Terminate the given session.

    SESSID: session ID given/generated when creating the session.
    """
    if len(session_names) == 0:
        print_warn('Specify at least one session ID. Check usage with "-h" option.')
        sys.exit(1)
    print_wait('Terminating the session(s)...')
    with Session() as session:
        has_failure = False
        for session_name in session_names:
            try:
                compute_session = session.ComputeSession(session_name, owner)
                ret = compute_session.destroy(forced=forced)
            except BackendAPIError as e:
                print_error(e)
                if e.status == 404:
                    print_info(
                        'If you are an admin, use "-o" / "--owner" option '
                        'to terminate other user\'s session.')
                has_failure = True
            except Exception as e:
                print_error(e)
                has_failure = True
        else:
            if not has_failure:
                print_done('Done.')
            if stats:
                stats = ret.get('stats', None) if ret else None
                if stats:
                    print(_format_stats(stats))
                else:
                    print('Statistics is not available.')
        if has_failure:
            sys.exit(1)


@main.command()
@click.argument('session_names', metavar='SESSION_NAME', nargs=-1)
def restart(session_names):
    """
    Restart the given session.

    SESSID: session ID given/generated when creating the session.
    """
    if len(session_names) == 0:
        print_warn('Specify at least one session ID. Check usage with "-h" option.')
        sys.exit(1)
    print_wait('Restarting the session(s)...')
    with Session() as session:
        has_failure = False
        for session_name in session_names:
            try:
                compute_session = session.ComputeSession(session_name)
                compute_session.restart()
            except BackendAPIError as e:
                print_error(e)
                if e.status == 404:
                    print_info(
                        'If you are an admin, use "-o" / "--owner" option '
                        'to terminate other user\'s session.')
                has_failure = True
            except Exception as e:
                print_error(e)
                has_failure = True
        else:
            if not has_failure:
                print_done('Done.')
        if has_failure:
            sys.exit(1)


@main.command()
@click.argument('session_id', metavar='SESSID')
@click.pass_context
def info(ctx, session_id):
    """
    Show detailed information for a running compute session.
    This is an alias of the "admin session <sess_id>" command.

    SESSID: session ID or its alias given when creating the session.
    """
    ctx.forward(cli_admin_session)


@main.command()
@click.argument('session_name_or_id', metavar='SESSION_ID_OR_NAME')
@click.option('-o', '--owner', '--owner-access-key', 'owner_access_key', metavar='ACCESS_KEY',
              help='Specify the owner of the target session explicitly.')
@click.option('--scope', type=click.Choice(['*', 'session', 'kernel']), default='*',
              help='Filter the events by kernel-specific ones or session-specific ones.')
def events(session_name_or_id, owner_access_key, scope):
    """
    Monitor the lifecycle events of a compute session.

    SESSID: session ID or its alias given when creating the session.
    """

    async def _run_events():
        async with AsyncSession() as session:
            try:
                session_id = uuid.UUID(session_name_or_id)
                compute_session = session.ComputeSession.from_session_id(session_id)
            except ValueError:
                compute_session = session.ComputeSession(session_name_or_id, owner_access_key)
            async with compute_session.listen_events(scope=scope) as response:
                async for ev in response:
                    print(click.style(ev.event, fg='cyan', bold=True), json.loads(ev.data))

    try:
        asyncio_run(_run_events())
    except Exception as e:
        print_error(e)



import click

from . import main
from .admin.sessions import sessions


@main.command()
@click.option('-s', '--status', default=None,
              type=click.Choice([
                  'PENDING', 'SCHEDULED',
                  'PREPARING', 'BUILDING', 'RUNNING', 'RESTARTING',
                  'RESIZING', 'SUSPENDED', 'TERMINATING',
                  'TERMINATED', 'ERROR', 'CANCELLED',
                  'ALL',  # special case
              ]),
              help='Filter by the given status')
@click.option('--name-only', is_flag=True, help='Display session names only.')
@click.option('--dead', is_flag=True,
              help='Filter only dead sessions. Ignores --status option.')
@click.option('--running', is_flag=True,
              help='Filter only scheduled and running sessions. Ignores --status option.')
@click.option('--detail', is_flag=True, help='Show more details using more columns.')
@click.option('-f', '--format', default=None,  help='Display only specified fields.')
@click.option('--filter', 'filter_', default=None, help='Set the query filter expression.')
@click.option('--order', default=None, help='Set the query ordering expression.')
@click.option('--plain', is_flag=True,
              help='Display the session list without decorative line drawings and the header.')
@click.pass_context
def ps(ctx, status, name_only, dead, running, detail, filter_, order, plain, format):
    '''
    Lists the current running compute sessions for the current keypair.
    This is an alias of the "admin sessions --status=RUNNING" command.
    '''
    ctx.forward(sessions)



import sys

import click
from tabulate import tabulate

from . import admin
from ..interaction import ask_yn
from ..pretty import print_error, print_info, print_fail
from ...session import Session


@admin.command()
@click.argument('gid', type=str)
def group(gid):
    '''
    Show the information about the given group.

    \b
    GID: Group ID.
    '''
    fields = [
        ('ID', 'id'),
        ('Name', 'name'),
        ('Domain', 'domain_name'),
        ('Description', 'description'),
        ('Active?', 'is_active'),
        ('Created At', 'created_at'),
        ('Total Resource Slots', 'total_resource_slots'),
        ('Allowed vFolder Hosts', 'allowed_vfolder_hosts'),
    ]
    with Session() as session:
        try:
            item = session.Group.detail(
                gid=gid,
                fields=(item[1] for item in fields),
            )
        except Exception as e:
            print_error(e)
            sys.exit(1)
        rows = []
        if item is None:
            print_fail('There is no such group.')
            sys.exit(1)
        for name, key in fields:
            if key in item:
                rows.append((name, item[key]))
        print(tabulate(rows, headers=('Field', 'Value')))


@admin.command()
@click.argument('name', type=str)
def groups_by_name(name: str):
    '''
    Show the information about the group(s) having the given name.
    Two or more groups in different domains may have the same name,
    so this may print out information of multiple groups if queried
    by a superadmin.

    \b
    name: Group name.
    '''
    fields = [
        ('ID', 'id'),
        ('Name', 'name'),
        ('Domain', 'domain_name'),
        ('Description', 'description'),
        ('Active?', 'is_active'),
        ('Created At', 'created_at'),
        ('Total Resource Slots', 'total_resource_slots'),
        ('Allowed vFolder Hosts', 'allowed_vfolder_hosts'),
    ]
    with Session() as session:
        try:
            items = session.Group.from_name(
                name,
                fields=(item[1] for item in fields),
            )
        except Exception as e:
            print_error(e)
            sys.exit(1)
        rows = []
        if not items:
            print_fail('There is no such group.')
            sys.exit(1)
        show_splitter = (len(items) > 1)
        for item_idx, item in enumerate(items):
            if show_splitter and item_idx > 0:
                print("=" * 40)
            for name, key in fields:
                if key in item:
                    rows.append((name, item[key]))
            print(tabulate(rows, headers=('Field', 'Value')))


@admin.group(invoke_without_command=True)
@click.pass_context
@click.option('-d', '--domain-name', type=str, default=None,
              help='Domain name to list groups belongs to it.')
def groups(ctx, domain_name):
    '''
    List and manage groups.
    (admin privilege required)
    '''
    if ctx.invoked_subcommand is not None:
        return
    fields = [
        ('ID', 'id'),
        ('Name', 'name'),
        ('Domain', 'domain_name'),
        ('Description', 'description'),
        ('Active?', 'is_active'),
        ('Created At', 'created_at'),
        ('Total Resource Slots', 'total_resource_slots'),
        ('Allowed vFolder Hosts', 'allowed_vfolder_hosts'),
        ('Allowed scaling groups', 'scaling_groups'),
    ]
    with Session() as session:
        try:
            resp = session.Group.list(domain_name=domain_name,
                                      fields=(item[1] for item in fields))
        except Exception as e:
            print_error(e)
            sys.exit(1)
        if len(resp) < 1:
            print_fail('There is no group.')
            return
        fields = [field for field in fields if field[1] in resp[0]]
        print(tabulate((item.values() for item in resp),
                        headers=(item[0] for item in fields)))


@groups.command()
@click.argument('domain_name', type=str, metavar='DOMAIN_NAME')
@click.argument('name', type=str, metavar='NAME')
@click.option('-d', '--description', type=str, default='',
              help='Description of new group.')
@click.option('-i', '--inactive', is_flag=True,
              help='New group will be inactive.')
@click.option('--total-resource-slots', type=str, default='{}',
              help='Set total resource slots.')
@click.option('--allowed-vfolder-hosts', type=str, multiple=True,
              help='Allowed virtual folder hosts.')
def add(domain_name, name, description, inactive, total_resource_slots,
        allowed_vfolder_hosts):
    '''
    Add new group. A group must belong to a domain, so DOMAIN_NAME should be provided.

    \b
    DOMAIN_NAME: Name of the domain where new group belongs to.
    NAME: Name of new group.
    '''
    with Session() as session:
        try:
            data = session.Group.create(
                domain_name, name,
                description=description,
                is_active=not inactive,
                total_resource_slots=total_resource_slots,
                allowed_vfolder_hosts=allowed_vfolder_hosts,
            )
        except Exception as e:
            print_error(e)
            sys.exit(1)
        if not data['ok']:
            print_fail('Group creation has failed: {0}'.format(data['msg']))
            sys.exit(1)
        item = data['group']
        print('Group name {0} is created in domain {1}.'.format(item['name'], item['domain_name']))


@groups.command()
@click.argument('gid', type=str, metavar='GROUP_ID')
@click.option('-n', '--name', type=str, help='New name of the group')
@click.option('-d', '--description', type=str, help='Description of the group')
@click.option('--is-active', type=bool, help='Set group inactive.')
@click.option('--total-resource-slots', type=str, help='Update total resource slots.')
@click.option('--allowed-vfolder-hosts', type=str, multiple=True,
              help='Allowed virtual folder hosts.')
def update(gid, name, description, is_active, total_resource_slots,
           allowed_vfolder_hosts):
    '''
    Update an existing group. Domain name is not necessary since group ID is unique.

    GROUP_ID: Group ID to update.
    '''
    with Session() as session:
        try:
            data = session.Group.update(
                gid,
                name=name,
                description=description,
                is_active=is_active,
                total_resource_slots=total_resource_slots,
                allowed_vfolder_hosts=allowed_vfolder_hosts,
            )
        except Exception as e:
            print_error(e)
            sys.exit(1)
        if not data['ok']:
            print_fail('Group update has failed: {0}'.format(data['msg']))
            sys.exit(1)
        print('Group {0} is updated.'.format(gid))


@groups.command()
@click.argument('gid', type=str, metavar='GROUP_ID')
def delete(gid):
    """
    Inactivates the existing group. Does not actually delete it for safety.

    GROUP_ID: Group ID to inactivate.
    """
    with Session() as session:
        try:
            data = session.Group.delete(gid)
        except Exception as e:
            print_error(e)
            sys.exit(1)
        if not data['ok']:
            print_fail('Group inactivation has failed: {0}'.format(data['msg']))
            sys.exit(1)
        print('Group is inactivated: ' + gid + '.')


@groups.command()
@click.argument('gid', type=str, metavar='GROUP_ID')
def purge(gid):
    """
    Delete the existing group. This action cannot be undone.

    GROUP_ID: Group ID to inactivate.
    """
    with Session() as session:
        try:
            if not ask_yn():
                print_info('Cancelled')
                sys.exit(1)
            data = session.Group.purge(gid)
        except Exception as e:
            print_error(e)
            sys.exit(1)
        if not data['ok']:
            print_fail('Group deletion has failed: {0}'.format(data['msg']))
            sys.exit(1)
        print('Group is deleted: ' + gid + '.')


@groups.command()
@click.argument('gid', type=str, metavar='GROUP_ID')
@click.argument('user_uuids', type=str, metavar='USER_UUIDS', nargs=-1)
def add_users(gid, user_uuids):
    '''
    Add users to a group.

    \b
    GROUP_ID: Group ID where users will be belong to.
    USER_UUIDS: List of users' uuids to be added to the group.
    '''
    with Session() as session:
        try:
            data = session.Group.add_users(gid, user_uuids)
        except Exception as e:
            print_error(e)
            sys.exit(1)
        if not data['ok']:
            print_fail('Error on adding users to group: {0}'.format(data['msg']))
            sys.exit(1)
        print('Users are added to the group')


@groups.command()
@click.argument('gid', type=str, metavar='GROUP_ID')
@click.argument('user_uuids', type=str, metavar='USER_UUIDS', nargs=-1)
def remove_users(gid, user_uuids):
    '''
    Remove users from a group.

    \b
    GROUP_ID: Group ID where users currently belong to.
    USER_UUIDS: List of users' uuids to be removed from the group.
    '''
    with Session() as session:
        try:
            data = session.Group.remove_users(gid, user_uuids)
        except Exception as e:
            print_error(e)
            sys.exit(1)
        if not data['ok']:
            print_fail('Error on removing users to group: {0}'.format(data['msg']))
            sys.exit(1)
        print('Users are removed from the group')



import sys

import click
import humanize
from tabulate import tabulate

from . import admin
from ..pagination import (
    get_preferred_page_size,
    echo_via_pager,
    tabulate_items,
)
from ..pretty import print_error
from ...session import Session
from ...exceptions import NoItems


@admin.group(invoke_without_command=True)
@click.pass_context
@click.option('--access-key', type=str, default=None,
              help='Get vfolders for the given access key '
                   '(only works if you are a super-admin)')
@click.option('-g', '--group', type=str, default=None,
              help='Filter by group ID.')
@click.option('--filter', 'filter_', default=None,
              help='Set the query filter expression.')
@click.option('--order', default=None,
              help='Set the query ordering expression.')
def vfolders(ctx, access_key, group, filter_, order):
    '''
    List and manage virtual folders.
    '''
    if ctx.invoked_subcommand is not None:
        return

    fields = [
        ('Name', 'name'),
        ('Host', 'host'),
        ('Group', 'group'),
        ('Creator', 'creator'),
        ('Permission', 'permission'),
        ('Usage Mode', 'usage_mode'),
        ('Ownership Type', 'ownership_type'),
        ('Created At', 'created_at'),
        ('Last Used', 'last_used'),
        ('Max Size', 'max_size'),
    ]

    try:
        with Session() as session:
            page_size = get_preferred_page_size()
            try:
                items = session.VFolder.paginated_list(
                    group, access_key,
                    fields=[f[1] for f in fields],
                    page_size=page_size,
                    filter=filter_,
                    order=order,
                )
                echo_via_pager(
                    tabulate_items(items, fields)
                )
            except NoItems:
                print("There are no matching vfolders.")
    except Exception as e:
        print_error(e)
        sys.exit(1)


@vfolders.command()
def list_hosts():
    '''
    List all mounted hosts from virtual folder root.
    (superadmin privilege required)
    '''
    with Session() as session:
        try:
            resp = session.VFolder.list_all_hosts()
            print("Default vfolder host: {}".format(resp['default']))
            print("Mounted hosts: {}".format(', '.join(resp['allowed'])))
        except Exception as e:
            print_error(e)
            sys.exit(1)


@vfolders.command()
@click.argument('vfolder_host')
def perf_metric(vfolder_host):
    '''
    Show the performance statistics of a vfolder host.
    (superadmin privilege required)

    A vfolder host consists of a string of the storage proxy name and the volume name
    separated by a colon. (e.g., "local:volume1")
    '''
    with Session() as session:
        try:
            resp = session.VFolder.get_performance_metric(vfolder_host)
            print(tabulate(
                [(k, humanize.naturalsize(v, binary=True) if 'bytes' in k else f"{v:.2f}")
                 for k, v in resp['metric'].items()],
                headers=('Key', 'Value'),
            ))
        except Exception as e:
            print_error(e)
            sys.exit(1)


@vfolders.command()
@click.option('-a', '--agent-id', type=str, default=None,
              help='Target agent to fetch fstab contents.')
def get_fstab_contents(agent_id):
    '''
    Get contents of fstab file from a node.
    (superadmin privilege required)

    If agent-id is not specified, manager's fstab contents will be returned.
    '''
    with Session() as session:
        try:
            resp = session.VFolder.get_fstab_contents(agent_id)
        except Exception as e:
            print_error(e)
            sys.exit(1)
        print(resp)


@vfolders.command()
def list_mounts():
    '''
    List all mounted hosts in virtual folder root.
    (superadmin privilege required)
    '''
    with Session() as session:
        try:
            resp = session.VFolder.list_mounts()
        except Exception as e:
            print_error(e)
            sys.exit(1)
        print('manager')
        for k, v in resp['manager'].items():
            print(' ', k, ':', v)
        print('\nagents')
        for aid, data in resp['agents'].items():
            print(' ', aid)
            for k, v in data.items():
                print('   ', k, ':', v)


@vfolders.command()
@click.argument('fs-location', type=str)
@click.argument('name', type=str)
@click.option('-o', '--options', type=str, default=None, help='Mount options.')
@click.option('--edit-fstab', is_flag=True,
              help='Edit fstab file to mount permanently.')
def mount_host(fs_location, name, options, edit_fstab):
    '''
    Mount a host in virtual folder root.
    (superadmin privilege required)

    \b
    FS-LOCATION: Location of file system to be mounted.
    NAME: Name of mounted host.
    '''
    with Session() as session:
        try:
            resp = session.VFolder.mount_host(name, fs_location, options, edit_fstab)
        except Exception as e:
            print_error(e)
            sys.exit(1)
        print('manager')
        for k, v in resp['manager'].items():
            print(' ', k, ':', v)
        print('agents')
        for aid, data in resp['agents'].items():
            print(' ', aid)
            for k, v in data.items():
                print('   ', k, ':', v)


@vfolders.command()
@click.argument('name', type=str)
@click.option('--edit-fstab', is_flag=True,
              help='Edit fstab file to mount permanently.')
def umount_host(name, edit_fstab):
    '''
    Unmount a host from virtual folder root.
    (superadmin privilege required)

    \b
    NAME: Name of mounted host.
    '''
    with Session() as session:
        try:
            resp = session.VFolder.umount_host(name, edit_fstab)
        except Exception as e:
            print_error(e)
            sys.exit(1)
        print('manager')
        for k, v in resp['manager'].items():
            print(' ', k, ':', v)
        print('agents')
        for aid, data in resp['agents'].items():
            print(' ', aid)
            for k, v in data.items():
                print('   ', k, ':', v)



import warnings

import click

from .. import __version__
from ..config import APIConfig, set_config
from ai.backend.cli.extensions import ExtendedCommandGroup


@click.group(
    cls=ExtendedCommandGroup,
    context_settings={
        'help_option_names': ['-h', '--help'],
    },
)
@click.option('--skip-sslcert-validation',
              help='Skip SSL certificate validation for all API requests.',
              is_flag=True)
@click.version_option(version=__version__)
def main(skip_sslcert_validation):
    """
    Backend.AI command line interface.
    """
    from .announcement import announce
    config = APIConfig(
        skip_sslcert_validation=skip_sslcert_validation,
        announcement_handler=announce,
    )
    set_config(config)

    from .pretty import show_warning
    warnings.showwarning = show_warning


def _attach_command():
    from . import admin, config, app, files, logs, manager, proxy, ps, run # noqa
    from . import ssh # noqa
    from . import vfolder       # noqa
    from . import session_template  # noqa
    from . import dotfile  # noqa
    from . import server_log  # noqa


_attach_command()



import json
import sys

import click
from tabulate import tabulate

from . import admin
from ...session import Session
from ..pretty import print_error
from ..pagination import (
    get_preferred_page_size,
    echo_via_pager,
    tabulate_items,
)
from ..utils import format_nested_dicts, format_value
from ...exceptions import NoItems


@admin.command()
@click.argument('vfolder_host')
def storage(vfolder_host):
    """
    Show the information about the given storage volume.
    (super-admin privilege required)
    """
    fields = [
        ('ID', 'id'),
        ('Backend', 'backend'),
        ('Capabilities', 'capabilities'),
        ('Path', 'path'),
        ('FS prefix', 'fsprefix'),
        ('Hardware Metadata', 'hardware_metadata'),
        ('Perf. Metric', 'performance_metric'),
    ]
    with Session() as session:
        try:
            resp = session.Storage.detail(
                vfolder_host=vfolder_host,
                fields=(item[1] for item in fields))
        except Exception as e:
            print_error(e)
            sys.exit(1)
        rows = []
        for name, key in fields:
            if key in resp:
                if key in ('hardware_metadata', 'performance_metric'):
                    rows.append((name, format_nested_dicts(json.loads(resp[key]))))
                else:
                    rows.append((name, format_value(resp[key])))
        print(tabulate(rows, headers=('Field', 'Value')))


@admin.command()
@click.option('--filter', 'filter_', default=None,
              help='Set the query filter expression.')
@click.option('--order', default=None,
              help='Set the query ordering expression.')
def storage_list(filter_, order):
    """
    List storage volumes.
    (super-admin privilege required)
    """
    fields = [
        ('ID', 'id'),
        ('Backend', 'backend'),
        ('Capabilities', 'capabilities'),
    ]
    try:
        with Session() as session:
            page_size = get_preferred_page_size()
            try:
                items = session.Storage.paginated_list(
                    fields=[f[1] for f in fields],
                    page_size=page_size,
                    filter=filter_,
                    order=order,
                )
                echo_via_pager(
                    tabulate_items(items, fields)
                )
            except NoItems:
                print("There are no storage volumes.")
    except Exception as e:
        print_error(e)
        sys.exit(1)



import sys

import click
from tabulate import tabulate

from . import admin
from ...session import Session
from ..interaction import ask_yn
from ..pretty import print_error, print_info, print_fail
from ..pagination import (
    get_preferred_page_size,
    echo_via_pager,
    tabulate_items,
)
from ...exceptions import NoItems


@admin.command()
@click.option('-e', '--email', type=str, default=None,
              help='Email of a user to display.')
def user(email):
    '''
    Show the information about the given user by email. If email is not give,
    requester's information will be displayed.
    '''
    fields = [
        ('UUID', 'uuid'),
        ('Username', 'username'),
        ('Role', 'role'),
        ('Email', 'email'),
        ('Name', 'full_name'),
        ('Need Password Change', 'need_password_change'),
        ('Status', 'status'),
        ('Status Info', 'status_info'),
        ('Created At', 'created_at'),
        ('Domain Name', 'domain_name'),
        ('Groups', 'groups { id name }'),
    ]
    with Session() as session:
        try:
            resp = session.User.detail(email=email, fields=(item[1] for item in fields))
        except Exception as e:
            print_error(e)
            sys.exit(1)
        rows = []
        if resp is None:
            print('There is no such user.')
            sys.exit(1)
        for name, key in fields:
            if key.startswith('groups '):
                group_list = [f"{g['name']} ({g['id']})" for g in resp['groups']]
                rows.append((name, ",\n".join(group_list)))
            else:
                rows.append((name, resp[key]))
        print(tabulate(rows, headers=('Field', 'Value')))


@admin.group(invoke_without_command=True)
@click.pass_context
@click.option('-s', '--status', type=str, default=None,
              help='Filter users in a specific state (active, inactive, deleted, before-verification).')
@click.option('-g', '--group', type=str, default=None,
              help='Filter by group ID.')
@click.option('--filter', 'filter_', default=None,
              help='Set the query filter expression.')
@click.option('--order', default=None,
              help='Set the query ordering expression.')
def users(ctx, status, group, filter_, order) -> None:
    '''
    List and manage users.
    (admin privilege required)
    '''
    if ctx.invoked_subcommand is not None:
        return
    fields = [
        ('UUID', 'uuid'),
        ('Username', 'username'),
        ('Role', 'role'),
        ('Email', 'email'),
        ('Name', 'full_name'),
        ('Need Password Change', 'need_password_change'),
        ('Status', 'status'),
        ('Status Info', 'status_info'),
        ('Created At', 'created_at'),
        ('Domain Name', 'domain_name'),
        ('Groups', 'groups { id name }'),
    ]

    def format_item(item):
        group_list = [g['name'] for g in item['groups']]
        item['groups'] = ", ".join(group_list)

    try:
        with Session() as session:
            page_size = get_preferred_page_size()
            try:
                items = session.User.paginated_list(
                    status, group,
                    fields=[f[1] for f in fields],
                    page_size=page_size,
                    filter=filter_,
                    order=order,
                )
                echo_via_pager(
                    tabulate_items(items, fields,
                                   item_formatter=format_item)
                )
            except NoItems:
                print("There are no matching users.")
    except Exception as e:
        print_error(e)
        sys.exit(1)


@users.command()
@click.argument('domain_name', type=str, metavar='DOMAIN_NAME')
@click.argument('email', type=str, metavar='EMAIL')
@click.argument('password', type=str, metavar='PASSWORD')
@click.option('-u', '--username', type=str, default='', help='Username.')
@click.option('-n', '--full-name', type=str, default='', help='Full name.')
@click.option('-r', '--role', type=str, default='user',
              help='Role of the user. One of (admin, user, monitor).')
@click.option('-s', '--status', type=str, default='active',
              help='Account status. One of (active, inactive, deleted, before-verification).')
@click.option('--need-password-change', is_flag=True,
              help='Flag indicate that user needs to change password. '
                   'Useful when admin manually create password.')
@click.option('--description', type=str, default='', help='Description of the user.')
def add(domain_name, email, password, username, full_name, role, status,
        need_password_change, description):
    '''
    Add new user. A user must belong to a domain, so DOMAIN_NAME should be provided.

    \b
    DOMAIN_NAME: Name of the domain where new user belongs to.
    EMAIL: Email of new user.
    PASSWORD: Password of new user.
    '''
    with Session() as session:
        try:
            data = session.User.create(
                domain_name, email, password,
                username=username, full_name=full_name, role=role,
                status=status,
                need_password_change=need_password_change,
                description=description,
            )
        except Exception as e:
            print_error(e)
            sys.exit(1)
        if not data['ok']:
            print_fail('User creation has failed: {0}'.format(data['msg']))
            sys.exit(1)
        item = data['user']
        print('User {0} is created in domain {1}.'.format(item['email'], item['domain_name']))


@users.command()
@click.argument('email', type=str, metavar='EMAIL')
@click.option('-p', '--password', type=str, help='Password.')
@click.option('-u', '--username', type=str, help='Username.')
@click.option('-n', '--full-name', type=str, help='Full name.')
@click.option('-d', '--domain-name', type=str, help='Domain name.')
@click.option('-r', '--role', type=str, default='user',
              help='Role of the user. One of (admin, user, monitor).')
@click.option('-s', '--status', type=str,
              help='Account status. One of (active, inactive, deleted, before-verification).')
@click.option('--need-password-change', is_flag=True,
              help='Flag indicate that user needs to change password. '
                   'Useful when admin manually create password.')
@click.option('--description', type=str, default='', help='Description of the user.')
def update(email, password, username, full_name, domain_name, role, status,
           need_password_change, description):
    '''
    Update an existing user.

    EMAIL: Email of user to update.
    '''
    with Session() as session:
        try:
            data = session.User.update(
                email,
                password=password, username=username, full_name=full_name,
                domain_name=domain_name,
                role=role, status=status, need_password_change=need_password_change,
                description=description,
            )
        except Exception as e:
            print_error(e)
            sys.exit(1)
        if not data['ok']:
            print_fail('User update has failed: {0}'.format(data['msg']))
            sys.exit(1)
        print('User {0} is updated.'.format(email))


@users.command()
@click.argument('email', type=str, metavar='EMAIL')
def delete(email):
    """
    Inactivate an existing user.

    EMAIL: Email of user to inactivate.
    """
    with Session() as session:
        try:
            data = session.User.delete(email)
        except Exception as e:
            print_error(e)
            sys.exit(1)
        if not data['ok']:
            print_fail('User inactivation has failed: {0}'.format(data['msg']))
            sys.exit(1)
        print('User is inactivated: ' + email + '.')


@users.command()
@click.argument('email', type=str, metavar='EMAIL')
@click.option('--purge-shared-vfolders', is_flag=True, default=False,
              help='Delete user\'s all virtual folders. '
                   'If False, shared folders will not be deleted '
                   'and migrated the ownership to the requested admin.')
def purge(email, purge_shared_vfolders):
    """
    Delete an existing user. This action cannot be undone.

    NAME: Name of a domain to delete.
    """
    with Session() as session:
        try:
            if not ask_yn():
                print_info('Cancelled')
                sys.exit(1)
            data = session.User.purge(email, purge_shared_vfolders)
        except Exception as e:
            print_error(e)
            sys.exit(1)
        if not data['ok']:
            print_fail('User deletion has failed: {0}'.format(data['msg']))
            sys.exit(1)
        print('User is deleted: ' + email + '.')



from collections import defaultdict
import functools
import sys
import textwrap
from typing import (
    Any,
    Dict,
    Mapping,
    Sequence,
)
import uuid

import click

from . import admin
from ...session import Session
from ...versioning import get_naming, apply_version_aware_fields
from ..pretty import print_error, print_fail
from ..pagination import (
    get_preferred_page_size,
    echo_via_pager,
    tabulate_items,
)
from ..utils import format_multiline, format_stats
from ...exceptions import NoItems


SessionItem = Dict[str, Any]


# Lets say formattable options are:
field_names = {
    'name':             ('Session Name',
                         lambda api_session: get_naming(api_session.api_version, 'name_gql_field')),
    'type':             ('Type',
                         lambda api_session: get_naming(api_session.api_version, 'type_gql_field')),
    'kernel_id':        ('Kernel/Task ID', 'id'),
    'session_id':       ('Session ID', 'session_id'),
    'status':           ('Status', 'status'),
    'status_info':      ('Status Info', 'status_info'),
    'status_data':      ('Status Data', 'status_data'),
    'created_at':       ('Created At', 'created_at'),
    'terminated_at':    ('Terminated At', 'terminated_at'),
    'last_updated':     ('Last updated', 'status_changed'),
    'result':           ('Result', 'result'),
    'group':            ('Project/Group', 'group_name'),
    'owner':            ('Owner', 'access_key'),
    'image':            ('Image', 'image'),
    'tag':              ('Tag', 'tag'),
    'occupied_slots':   ('Occupied Resource', 'occupied_slots'),
    'cluster_hostname': ('Hostname', 'cluster_hostname'),
}

field_names_legacy = {
    'used_memory':     ('Used Memory (MiB)', 'mem_cur_bytes'),
    'max_used_memory': ('Max Used Memory (MiB)', 'mem_max_bytes'),
    'cpu_using':       ('CPU Using (%)', 'cpu_using'),
}


def transform_fields(item: SessionItem, *, in_row: bool = True) -> SessionItem:
    if 'mem_cur_bytes' in item:
        item['mem_cur_bytes'] = round(item['mem_cur_bytes'] / 2 ** 20, 1)
    if 'mem_max_bytes' in item:
        item['mem_max_bytes'] = round(item['mem_max_bytes'] / 2 ** 20, 1)
    return item


@admin.command()
@click.option('-s', '--status', default=None,
              type=click.Choice([
                  'PENDING', 'SCHEDULED',
                  'PREPARING', 'BUILDING', 'RUNNING', 'RESTARTING',
                  'RESIZING', 'SUSPENDED', 'TERMINATING',
                  'TERMINATED', 'ERROR', 'CANCELLED',
                  'ALL',  # special case
              ]),
              help='Filter by the given status')
@click.option('--access-key', type=str, default=None,
              help='Get sessions for a specific access key '
                   '(only works if you are a super-admin)')
@click.option('--name-only', is_flag=True, help='Display session names only.')
@click.option('--dead', is_flag=True,
              help='Filter only dead sessions. Ignores --status option.')
@click.option('--running', is_flag=True,
              help='Filter only scheduled and running sessions. Ignores --status option.')
@click.option('--detail', is_flag=True, help='Show more details using more columns.')
@click.option('-f', '--format', default=None, help='Display only specified fields.')
@click.option('--filter', 'filter_', default=None, help='Set the query filter expression.')
@click.option('--order', default=None, help='Set the query ordering expression.')
@click.option('--plain', is_flag=True,
              help='Display the session list without decorative line drawings and the header.')
def sessions(status, access_key, name_only, dead, running, detail, filter_, order, plain, format):
    '''
    List and manage compute sessions.
    '''
    fields = []
    with Session() as session:
        is_admin = session.KeyPair(session.config.access_key).info()['is_admin']
        try:
            name_key = get_naming(session.api_version, 'name_gql_field')
            fields.append(field_names['name'])
            if is_admin:
                fields.append(field_names['owner'])
        except Exception as e:
            print_error(e)
            sys.exit(1)
        if name_only:
            pass
        elif format is not None:
            options = format.split(',')
            for opt in options:
                if opt not in field_names:
                    print_fail(f'There is no such format option: {opt}')
                    sys.exit(1)
            fields = [
                field_names[opt] for opt in options
            ]
        else:
            if session.api_version[0] >= 6:
                fields.append(field_names['session_id'])
            fields.extend([
                field_names['group'],
                field_names['kernel_id'],
                field_names['image'],
                field_names['type'],
                field_names['status'],
                field_names['status_info'],
                field_names['last_updated'],
                field_names['result'],
            ])
            if detail:
                fields.extend([
                    field_names['tag'],
                    field_names['created_at'],
                    field_names['occupied_slots'],
                ])
                if session.api_version[0] < 5:
                    fields.extend([
                        field_names_legacy['used_memory'],
                        field_names_legacy['max_used_memory'],
                        field_names_legacy['cpu_using'],
                    ])

    no_match_name = None
    if status is None:
        status = ('PENDING,SCHEDULED,PREPARING,PULLING,RUNNING,RESTARTING,TERMINATING,'
                  'RESIZING,SUSPENDED,ERROR')
        no_match_name = 'active'
    if running:
        status = 'PREPARING,PULLING,RUNNING'
        no_match_name = 'running'
    if dead:
        status = 'CANCELLED,TERMINATED'
        no_match_name = 'dead'
    if status == 'ALL':
        status = ('PENDING,SCHEDULED,PREPARING,PULLING,RUNNING,RESTARTING,TERMINATING,'
                  'RESIZING,SUSPENDED,ERROR,'
                  'CANCELLED,TERMINATED')
        no_match_name = 'in any status'
    if no_match_name is None:
        no_match_name = status.lower()

    try:
        with Session() as session:
            fields = apply_version_aware_fields(session, fields)
            # let the page size be same to the terminal height.
            page_size = get_preferred_page_size()
            try:
                items = session.ComputeSession.paginated_list(
                    status, access_key,
                    fields=[f[1] for f in fields],
                    page_size=page_size,
                    filter=filter_,
                    order=order,
                )
                if name_only:
                    echo_via_pager(
                        (f"{item[name_key]}\n" for item in items)
                    )
                else:
                    echo_via_pager(
                        tabulate_items(items, fields,
                                       item_formatter=transform_fields)
                    )
            except NoItems:
                print("There are no matching sessions.")
    except Exception as e:
        print_error(e)
        sys.exit(1)


def format_containers(containers: Sequence[Mapping[str, Any]], indent='') -> str:

    if len(containers) == 0:
        text = "- (There are no sub-containers belonging to the session)"
    else:
        text = ""
        for cinfo in containers:
            text += "\n".join((
                f"+ {cinfo['id']}",
                *(f"  - {k + ': ':18s}{format_multiline(v, 22)}"
                  for k, v in cinfo.items()
                  if k not in ('id', 'live_stat', 'last_stat')),
                f"  + live_stat: {format_stats(cinfo['live_stat'], indent='    ')}",
                f"  + last_stat: {format_stats(cinfo['last_stat'], indent='    ')}",
            )) + "\n"
    return "\n" + textwrap.indent(text, indent)


def format_dependencies(dependencies: Sequence[Mapping[str, Any]], indent='') -> str:
    if len(dependencies) == 0:
        text = "- (There are no dependency tasks)"
    else:
        text = ""
        for dinfo in dependencies:
            text += "\n".join(
                (f"+ {dinfo['name']} ({dinfo['id']})",
                *(f"  - {k + ': ':18s}{v}" for k, v in dinfo.items() if k not in ('name', 'id'))),
            )
    return "\n" + textwrap.indent(text, indent)


@admin.command()
@click.argument('session_id', metavar='SESSID')
def session(session_id):
    '''
    Show detailed information for a running compute session.
    '''
    with Session() as session_:
        fields = [
            ('Session Name', lambda api_session: get_naming(
                api_session.api_version, 'name_gql_field',
            )),
        ]
        if session_.api_version[0] >= 6:
            fields.append(field_names['session_id'])
            fields.append(field_names['kernel_id'])
        fields.extend([
            field_names['image'],
        ])
        if session_.api_version >= (4, '20181215'):
            fields.append(field_names['tag'])
        fields.extend([
            field_names['created_at'],
            field_names['terminated_at'],
            field_names['status'],
            field_names['status_info'],
            field_names['status_data'],
            field_names['occupied_slots'],
        ])
        fields = apply_version_aware_fields(session_, fields)
        field_formatters = defaultdict(lambda: str)
        field_formatters['last_stat'] = format_stats
        field_formatters['containers'] = functools.partial(format_containers, indent='  ')
        field_formatters['dependencies'] = functools.partial(format_dependencies, indent='  ')
        if session_.api_version[0] < 5:
            # In API v4 or older, we can only query a currently running session
            # using its user-defined alias name.
            fields.append(('Last Stats', 'last_stat'))
            q = 'query($name: String!) {' \
                '  compute_session(sess_id: $name) { $fields }' \
                '}'
            v = {'name': session_id}
        else:
            # In API v5 or later, we can query any compute session both in the history
            # and currently running using its UUID.
            # NOTE: Partial ID/alias matching is supported in the REST API only.
            if session_.api_version[0] >= 6:
                fields.append((
                    'Containers',
                    'containers {'
                    ' id cluster_role cluster_idx cluster_hostname '
                    ' agent status status_info status_data status_changed '
                    ' occupied_slots live_stat last_stat '
                    '}',
                ))
            else:
                fields.append((
                    'Containers',
                    'containers {'
                    ' id role agent status status_info status_changed '
                    ' occupied_slots live_stat last_stat '
                    '}',
                ))
            fields.append((
                'Dependencies',
                'dependencies { name id status status_info status_changed }',
            ))
            q = 'query($id: UUID!) {' \
                '  compute_session(id: $id) {' \
                '    $fields' \
                '  }' \
                '}'
            try:
                uuid.UUID(session_id)
            except ValueError:
                print_fail("In API v5 or later, the session ID must be given in the UUID format.")
                sys.exit(1)
            v = {'id': session_id}
        q = q.replace('$fields', ' '.join(item[1] for item in fields))
        try:
            resp = session_.Admin.query(q, v)
        except Exception as e:
            print_error(e)
            sys.exit(1)
        if resp['compute_session'] is None:
            if session_.api_version[0] < 5:
                print_fail('There is no such running compute session.')
            else:
                print_fail('There is no such compute session.')
            sys.exit(1)
        transform_fields(resp['compute_session'], in_row=False)
        for i, (key, value) in enumerate(resp['compute_session'].items()):
            fmt = field_formatters[key]
            print(f"{fields[i][0] + ': ':20s}{fmt(value)}")
